#include <dbus/dbus-glib.h>

#include "accounts.h"
#include "common.h"
#include "signals-marshal.h"

enum {
  INVALID_ACCOUNT,
  UNREGISTERED,
  REGISTERED
};

/* Beginning of a classic GObject declaration */

typedef struct Accounts Accounts;
typedef struct AccountsClass AccountsClass;

GType accounts_get_type (void);

struct Accounts
{
  GtkFrame parent;

  DBusGProxy *dbus_object;
  GtkWidget *tree;
};

struct AccountsClass
{
  GtkFrameClass parent;
};

#define ACCOUNTS_TYPE (accounts_get_type ())
#define ACCOUNTS(object) (G_TYPE_CHECK_INSTANCE_CAST ((object), ACCOUNTS_TYPE, Accounts))
#define ACCOUNTS_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST ((klass), ACCOUNTS_TYPE, AccountsClass))
#define IS_ACCOUNTS(object) (G_TYPE_CHECK_INSTANCE_TYPE ((object), ACCOUNTS_TYPE))
#define IS_ACCOUNTS_CLASS(klass)   (G_TYPE_CHECK_CLASS_TYPE ((klass), ACCOUNTS_TYPE))

G_DEFINE_TYPE(Accounts, accounts, GTK_TYPE_FRAME);

/* End of a classic GObject declaration */

/* Declaration of the methods of this object */
static void accounts_populate (Accounts *self);

static void accounts_flush (Accounts *self);

static void accounts_update (Accounts *self,
			     const gchar *name,
			     guint state);
static void accounts_register (Accounts *self,
			       const gchar *name);

static void accounts_unregister (Accounts *self,
				 const gchar *name);

/* Declaration of some helper functions */
static void register_button_callback (GtkWidget *widget, gpointer data);
static void unregister_button_callback (GtkWidget *widget, gpointer data);
static void update_button_callback (GtkWidget *widget, gpointer data);

static void account_register_signal_handler (DBusGProxy *proxy,
					     const char *name,
					     guint state,
					     gpointer data);

/* Implementation */
static void
accounts_init (Accounts *self)
{
  DBusGConnection *bus = NULL;
  GError *error = NULL;
  GtkWidget *vbox = NULL;
  GtkWidget *hbox = NULL;
  GtkWidget *button = NULL;
  GtkListStore *store = NULL;
  GtkCellRenderer *renderer = NULL;
  GtkTreeViewColumn *column = NULL;

  bus = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
  if (bus == NULL) {

    print_error ("ACCOUNTS", "Couldn't connect to session bus", error);
    return;
  }

  self->dbus_object = dbus_g_proxy_new_for_name (bus,
						 "net.gnomemeeting.instance",
						 "/net/gnomemeeting/instance",
						 "net.gnomemeeting.accounts");

  if (self->dbus_object == NULL) {

    print_error ("ACCOUNTS", "Failed to get name owner", error);
    return;
  }
  dbus_g_proxy_add_signal (self->dbus_object, "AccountRegister",
			   G_TYPE_STRING, G_TYPE_UINT,
			   G_TYPE_INVALID);

  dbus_g_proxy_connect_signal (self->dbus_object, "AccountRegister",
			       G_CALLBACK (account_register_signal_handler),
                               (gpointer)self, NULL);

  store = gtk_list_store_new (2, G_TYPE_STRING, G_TYPE_BOOLEAN);
  self->tree = gtk_tree_view_new_with_model (GTK_TREE_MODEL (store));
  renderer = gtk_cell_renderer_text_new ();
  column = gtk_tree_view_column_new_with_attributes ("Name",
						     renderer,
						     "text", 0,
						     NULL);
  gtk_tree_view_append_column (GTK_TREE_VIEW (self->tree), column);

  renderer = gtk_cell_renderer_toggle_new ();
  column = gtk_tree_view_column_new_with_attributes ("Registered",
						     renderer,
						     "active", 1,
						     NULL);
  gtk_tree_view_append_column (GTK_TREE_VIEW (self->tree), column);

  hbox = gtk_hbox_new (FALSE, 5);

  button = gtk_button_new_with_label ("Register");
  gtk_container_add (GTK_CONTAINER (hbox), button);
  g_signal_connect (G_OBJECT (button), "clicked",
		    G_CALLBACK (register_button_callback), self);

  button = gtk_button_new_with_label ("Unregister");
  gtk_container_add (GTK_CONTAINER (hbox), button);
  g_signal_connect (G_OBJECT (button), "clicked",
		    G_CALLBACK (unregister_button_callback), self);

  button = gtk_button_new_with_label ("Update");
  gtk_container_add (GTK_CONTAINER (hbox), button);
  g_signal_connect (G_OBJECT (button), "clicked",
		    G_CALLBACK (update_button_callback), self);

  vbox = gtk_vbox_new (FALSE, 5);
  gtk_container_add (GTK_CONTAINER (vbox), self->tree);
  gtk_container_add (GTK_CONTAINER (vbox), hbox);

  gtk_container_add (GTK_CONTAINER (self), vbox);

}

static void
accounts_class_init (AccountsClass *klass)
{
  dbus_g_object_register_marshaller (signals_VOID__STRING_UINT,
				     G_TYPE_NONE,
				     G_TYPE_STRING, G_TYPE_UINT,
				     G_TYPE_INVALID);
}

static void
accounts_populate (Accounts *self)
{
  GError *error = NULL;
  char **accounts_list = NULL;
  char **ptr;

  if (!dbus_g_proxy_call (self->dbus_object, "GetAccountsList", &error,
                          G_TYPE_INVALID,
                          G_TYPE_STRV, &accounts_list,
			  G_TYPE_INVALID)) {

    print_error ("ACCOUNTS", "Couldn't GetAccountsList", error);
    return;
  }

  for (ptr = accounts_list; *ptr; ptr++) {

    dbus_g_proxy_call_no_reply (self->dbus_object,
				"ResignalAccountState",
				G_TYPE_STRING, *ptr,
				G_TYPE_INVALID);
  }

  g_strfreev (accounts_list);
}

static void accounts_flush (Accounts *self)
{
  GtkListStore *store = NULL;

  store = GTK_LIST_STORE (gtk_tree_view_get_model (GTK_TREE_VIEW(self->tree)));
  gtk_list_store_clear (store);
}

static void
accounts_update (Accounts *self,
		 const gchar *name,
		 guint state)
{
  GtkTreeIter *iter = NULL;
  GtkListStore *store = NULL;

  store = GTK_LIST_STORE (gtk_tree_view_get_model (GTK_TREE_VIEW(self->tree)));

  iter = find_iter_for_key (name, 0, store);

  switch (state) {

  case INVALID_ACCOUNT:
    (void)gtk_list_store_remove (store, iter);
    break;
  case UNREGISTERED:
    gtk_list_store_set (store, iter,
			0, name,
			1, FALSE,
			-1);
    break;
  case REGISTERED:
    gtk_list_store_set (store, iter,
			0, name,
			1, TRUE,
			-1);
    break;
  default:
    print_error ("ACCOUNTS", "Received unknown state code !?", NULL);
  }

  gtk_tree_iter_free (iter);
}

static void
accounts_register (Accounts *self,
		   const gchar *name)
{
  GError *error = NULL;

  dbus_g_proxy_call_no_reply (self->dbus_object, "Register",
			      G_TYPE_STRING, name,
			      G_TYPE_INVALID);
}

static void
accounts_unregister (Accounts *self,
		     const gchar *name)
{
  GError *error = NULL;

  dbus_g_proxy_call_no_reply (self->dbus_object, "Unregister",
			      G_TYPE_STRING, name,
			      G_TYPE_INVALID);
}

static void
register_button_callback (GtkWidget *widget, gpointer data)
{
  Accounts *self = NULL;
  GtkTreeSelection *selection = NULL;
  GtkTreeModel *model = NULL;
  GList *paths = NULL;
  GList *ptr = NULL;
  GtkTreeIter iter;
  gchar *name = NULL;

  self = ACCOUNTS (data);
  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (self->tree));
  model = gtk_tree_view_get_model (GTK_TREE_VIEW (self->tree));
  paths = gtk_tree_selection_get_selected_rows (selection, &model);

  for (ptr = paths ; ptr != NULL; ptr = ptr->next) {

    if (gtk_tree_model_get_iter (model, &iter,(GtkTreePath *)ptr->data)) {

            gtk_tree_model_get (model, &iter, 0, &name, -1);
	    accounts_register (self, name);
	    g_free (name);
    }
  }

  g_list_foreach (paths, (GFunc)gtk_tree_path_free, NULL);
  g_list_free (paths);
}

static void
unregister_button_callback (GtkWidget *widget, gpointer data)
{
  Accounts *self = NULL;
  GtkTreeSelection *selection = NULL;
  GtkTreeModel *model = NULL;
  GList *paths = NULL;
  GList *ptr = NULL;
  GtkTreeIter iter;
  gchar *name = NULL;

  self = ACCOUNTS (data);
  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (self->tree));
  model = gtk_tree_view_get_model (GTK_TREE_VIEW (self->tree));
  paths = gtk_tree_selection_get_selected_rows (selection, &model);

  for (ptr = paths ; ptr != NULL; ptr = ptr->next) {

    if (gtk_tree_model_get_iter (model, &iter,(GtkTreePath *)ptr->data)) {

            gtk_tree_model_get (model, &iter, 0, &name, -1);
	    accounts_unregister (self, name);
	    g_free (name);
    }
  }

  g_list_foreach (paths, (GFunc)gtk_tree_path_free, NULL);
  g_list_free (paths);
}

static void
update_button_callback (GtkWidget *widget, gpointer data)
{
  Accounts *self = ACCOUNTS (data);

  accounts_flush (self);
  accounts_populate (self);
}

static void
account_register_signal_handler (DBusGProxy *proxy,
				 const char *name,
				 guint state,
				 gpointer data)
{
  accounts_update (ACCOUNTS (data), name, state);
}


/* implementation of the public functions */
GtkWidget *
accounts_new (void)
{
  Accounts *result = NULL;

  result = ACCOUNTS (gtk_object_new (ACCOUNTS_TYPE,
				     "label", "Accounts",
				     NULL));

  return GTK_WIDGET(result);
}
