#include <dbus/dbus-glib.h>

#include "calls.h"
#include "common.h"
#include "signals-marshal.h"

enum {
  INVALID_CALL,
  STANDBY,
  CALLING,
  CONNECTED,
  CALLED
};

static const char *states [] = {
  "INVALID_CALL",
  "Standby",
  "Calling",
  "Connected",
  "Called"
};

enum {
  COLUMN_TOKEN,
  COLUMN_STATE,
  COLUMN_NAME,
  COLUMN_CLIENT,
  COLUMN_URL,
  COLUMN_PROTOCOL,
  COLUMN_NUMBER
};

/* Beginning of a classic GObject declaration */

typedef struct Calls Calls;
typedef struct CallsClass CallsClass;

GType calls_get_type (void);

struct Calls
{
  GtkFrame parent;

  DBusGProxy *dbus_object;
  GtkWidget *tree;
};

struct CallsClass
{
  GtkFrameClass parent;
};

#define CALLS_TYPE (calls_get_type ())
#define CALLS(object) (G_TYPE_CHECK_INSTANCE_CAST ((object), CALLS_TYPE, Calls))
#define CALLS_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST ((klass), CALLS_TYPE, CallsClass))
#define IS_CALLS(object) (G_TYPE_CHECK_INSTANCE_TYPE ((object), CALLS_TYPE))
#define IS_CALLS_CLASS(klass)   (G_TYPE_CHECK_CLASS_TYPE ((klass), CALLS_TYPE))

G_DEFINE_TYPE(Calls, calls, GTK_TYPE_FRAME);

/* End of a classic GObject declaration */

/* Declaration of the methods of this object */
static void calls_populate (Calls *self);

static void calls_flush (Calls *self);

static void calls_update_state (Calls *self,
				const gchar *token,
				guint state);

static void calls_connect (Calls *self,
			   const gchar *url);

static void calls_disconnect (Calls *self,
			      const gchar *token);

static void calls_play_pause (Calls *self,
			      const gchar *token);

static void calls_transfer (Calls *self,
			    const gchar *token,
			    const gchar *url);

/* Declaration of some helper functions */
static void update_column (Calls *self, const gchar *token,
			   gint column, const gchar *info);

static void connect_entry_activate_callback (GtkWidget *widget, gpointer data);
static void transfer_entry_activate_callback (GtkWidget *widget, gpointer data);
static void disconnect_button_callback (GtkWidget *widget, gpointer data);
static void play_pause_button_callback (GtkWidget *widget, gpointer data);
static void update_button_callback (GtkWidget *widget, gpointer data);

static void state_changed_signal_handler (DBusGProxy *proxy,
					  const char *token,
					  guint state,
					  gpointer data);
static void name_info_signal_handler (DBusGProxy *proxy,
				      const char *token,
				      const char *name,
				      gpointer data);
static void client_info_signal_handler (DBusGProxy *proxy,
					const char *token,
					const char *client,
					gpointer data);
static void url_info_signal_handler (DBusGProxy *proxy,
				     const char *token,
				     const char *url,
				     gpointer data);
static void protocol_info_signal_handler (DBusGProxy *proxy,
					  const char *token,
					  const char *protocol,
					  gpointer data);

/* Implementation */
static void
calls_init (Calls *self)
{
  DBusGConnection *bus = NULL;
  GError *error = NULL;
  GtkWidget *vbox = NULL;
  GtkWidget *hbox = NULL;
  GtkListStore *store = NULL;
  GtkCellRenderer *renderer = NULL;
  GtkTreeViewColumn *column = NULL;
  GtkWidget *button = NULL;
  GtkWidget *transfer_hbox = NULL;
  GtkWidget *transfer_label = NULL;
  GtkWidget *transfer_entry = NULL;
  GtkWidget *connect_hbox = NULL;
  GtkWidget *connect_label = NULL;
  GtkWidget *connect_entry = NULL;

  bus = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
  if (bus == NULL) {

    print_error ("CALLS", "Couldn't connect to session bus", error);
    return;
  }

  self->dbus_object = dbus_g_proxy_new_for_name (bus,
						 "net.gnomemeeting.instance",
						 "/net/gnomemeeting/instance",
						 "net.gnomemeeting.calls");

  if (self->dbus_object == NULL) {

    print_error ("CALLS", "Failed to get name owner", error);
    return;
  }

  dbus_g_proxy_add_signal (self->dbus_object, "StateChanged",
			   G_TYPE_STRING, G_TYPE_UINT,
			   G_TYPE_INVALID);
  dbus_g_proxy_connect_signal (self->dbus_object, "StateChanged",
			       G_CALLBACK (state_changed_signal_handler),
                               (gpointer)self, NULL);

  dbus_g_proxy_add_signal (self->dbus_object, "NameInfo",
			   G_TYPE_STRING, G_TYPE_STRING,
			   G_TYPE_INVALID);
  dbus_g_proxy_connect_signal (self->dbus_object, "NameInfo",
			       G_CALLBACK (name_info_signal_handler),
                               (gpointer)self, NULL);

  dbus_g_proxy_add_signal (self->dbus_object, "ClientInfo",
			   G_TYPE_STRING, G_TYPE_STRING,
			   G_TYPE_INVALID);
  dbus_g_proxy_connect_signal (self->dbus_object, "ClientInfo",
			       G_CALLBACK (client_info_signal_handler),
                               (gpointer)self, NULL);

  dbus_g_proxy_add_signal (self->dbus_object, "UrlInfo",
			   G_TYPE_STRING, G_TYPE_STRING,
			   G_TYPE_INVALID);
  dbus_g_proxy_connect_signal (self->dbus_object, "UrlInfo",
			       G_CALLBACK (url_info_signal_handler),
                               (gpointer)self, NULL);

  dbus_g_proxy_add_signal (self->dbus_object, "ProtocolInfo",
			   G_TYPE_STRING, G_TYPE_STRING,
			   G_TYPE_INVALID);
  dbus_g_proxy_connect_signal (self->dbus_object, "ProtocolInfo",
			       G_CALLBACK (protocol_info_signal_handler),
                               (gpointer)self, NULL);

  store = gtk_list_store_new (COLUMN_NUMBER,
			      G_TYPE_STRING, /* COLUMN_TOKEN */
			      G_TYPE_STRING, /* COLUMN_STATE */
			      G_TYPE_STRING, /* COLUMN_NAME */
			      G_TYPE_STRING, /* COLUMN_CLIENT */
			      G_TYPE_STRING, /* COLUMN_URL */
			      G_TYPE_STRING);/* COLUMN_PROTOCOL */

  self->tree = gtk_tree_view_new_with_model (GTK_TREE_MODEL (store));

  renderer = gtk_cell_renderer_text_new ();
  column = gtk_tree_view_column_new_with_attributes ("State",
						     renderer,
						     "text", COLUMN_STATE,
						     NULL);
  gtk_tree_view_append_column (GTK_TREE_VIEW (self->tree), column);

  renderer = gtk_cell_renderer_text_new ();
  column = gtk_tree_view_column_new_with_attributes ("Name",
						     renderer,
						     "text", COLUMN_NAME,
						     NULL);
  gtk_tree_view_append_column (GTK_TREE_VIEW (self->tree), column);

  renderer = gtk_cell_renderer_text_new ();
  column = gtk_tree_view_column_new_with_attributes ("Client",
						     renderer,
						     "text", COLUMN_CLIENT,
						     NULL);
  gtk_tree_view_append_column (GTK_TREE_VIEW (self->tree), column);

  renderer = gtk_cell_renderer_text_new ();
  column = gtk_tree_view_column_new_with_attributes ("URL",
						     renderer,
						     "text", COLUMN_URL,
						     NULL);
  gtk_tree_view_append_column (GTK_TREE_VIEW (self->tree), column);

  renderer = gtk_cell_renderer_text_new ();
  column = gtk_tree_view_column_new_with_attributes ("Protocol",
						     renderer,
						     "text", COLUMN_PROTOCOL,
						     NULL);
  gtk_tree_view_append_column (GTK_TREE_VIEW (self->tree), column);

  hbox = gtk_hbox_new (FALSE, 5);

  button = gtk_button_new_with_label ("Disconnect");
  gtk_container_add (GTK_CONTAINER (hbox), button);
  g_signal_connect (G_OBJECT (button), "clicked",
		    G_CALLBACK (disconnect_button_callback), self);

  button = gtk_button_new_with_label ("Play/Pause");
  gtk_container_add (GTK_CONTAINER (hbox), button);
  g_signal_connect (G_OBJECT (button), "clicked",
		    G_CALLBACK (play_pause_button_callback), self);

  button = gtk_button_new_with_label ("Update");
  gtk_container_add (GTK_CONTAINER (hbox), button);
  g_signal_connect (G_OBJECT (button), "clicked",
		    G_CALLBACK (update_button_callback), self);

  connect_label = gtk_label_new ("Connect to: ");

  connect_entry = gtk_entry_new ();
  gtk_entry_set_text (GTK_ENTRY (connect_entry), "sip:");
  g_signal_connect (G_OBJECT (connect_entry), "activate",
		    G_CALLBACK (connect_entry_activate_callback), self);

  connect_hbox = gtk_hbox_new (FALSE, 5);
  gtk_container_add (GTK_CONTAINER (connect_hbox), connect_label);
  gtk_container_add (GTK_CONTAINER (connect_hbox), connect_entry);

  transfer_label = gtk_label_new ("Transfer selected to: ");

  transfer_entry = gtk_entry_new ();
  gtk_entry_set_text (GTK_ENTRY (transfer_entry), "sip:");
  g_signal_connect (G_OBJECT (transfer_entry), "activate",
		    G_CALLBACK (transfer_entry_activate_callback), self);

  transfer_hbox = gtk_hbox_new (FALSE, 5);
  gtk_container_add (GTK_CONTAINER (transfer_hbox), transfer_label);
  gtk_container_add (GTK_CONTAINER (transfer_hbox), transfer_entry);

  vbox = gtk_vbox_new (FALSE, 5);
  gtk_container_add (GTK_CONTAINER (vbox), self->tree);
  gtk_container_add (GTK_CONTAINER (vbox), hbox);
  gtk_container_add (GTK_CONTAINER (vbox), connect_hbox);
  gtk_container_add (GTK_CONTAINER (vbox), transfer_hbox);

  gtk_container_add (GTK_CONTAINER (self), vbox);

}

static void
calls_class_init (CallsClass *klass)
{
  dbus_g_object_register_marshaller (signals_VOID__STRING_UINT,
				     G_TYPE_NONE,
				     G_TYPE_STRING, G_TYPE_UINT,
				     G_TYPE_INVALID);
  dbus_g_object_register_marshaller (signals_VOID__STRING_STRING,
				     G_TYPE_NONE,
				     G_TYPE_STRING, G_TYPE_STRING,
				     G_TYPE_INVALID);
}

static void
calls_populate (Calls *self)
{
  GError *error = NULL;
  char **calls_list = NULL;
  char **ptr;

  if (!dbus_g_proxy_call (self->dbus_object, "GetCallsList", &error,
                          G_TYPE_INVALID,
                          G_TYPE_STRV, &calls_list,
			  G_TYPE_INVALID)) {

    print_error ("CALLS", "Couldn't GetCallsList", error);
    return;
  }

  for (ptr = calls_list; *ptr; ptr++) {

    dbus_g_proxy_call_no_reply (self->dbus_object,
				"ResignalCallInfo",
				G_TYPE_STRING, *ptr,
				G_TYPE_INVALID);
  }

  g_strfreev (calls_list);
}

static void calls_flush (Calls *self)
{
  GtkListStore *store = NULL;

  store = GTK_LIST_STORE (gtk_tree_view_get_model (GTK_TREE_VIEW(self->tree)));
  gtk_list_store_clear (store);
}

static void
calls_update_state (Calls *self,
		    const gchar *token,
		    guint state)
{
  GtkTreeIter *iter = NULL;
  GtkListStore *store = NULL;

  switch (state) {

  case INVALID_CALL:
    store =
      GTK_LIST_STORE (gtk_tree_view_get_model (GTK_TREE_VIEW(self->tree)));
    iter = find_iter_for_key (token, 0, store);
    (void)gtk_list_store_remove (store, iter);
    gtk_tree_iter_free (iter);
    break;
  case STANDBY:
  case CALLING:
  case CONNECTED:
  case CALLED:
    update_column (self, token, COLUMN_STATE, states[state]);
    break;
  default:
    print_error ("CALLS", "Received unknown state code !?", NULL);
  }

}

static void
calls_connect (Calls *self, const gchar *url)
{
  GError *error = NULL;

  dbus_g_proxy_call_no_reply (self->dbus_object, "Connect",
			      G_TYPE_STRING, url,
			      G_TYPE_INVALID);
}

static void
calls_disconnect (Calls *self,
		  const gchar *token)
{
  GError *error = NULL;

  dbus_g_proxy_call_no_reply (self->dbus_object, "Disconnect",
			      G_TYPE_STRING, token,
			      G_TYPE_INVALID);
}

static void
calls_play_pause (Calls *self,
		  const gchar *token)
{
  GError *error = NULL;

  dbus_g_proxy_call_no_reply (self->dbus_object, "PlayPause",
			      G_TYPE_STRING, token,
			      G_TYPE_INVALID);
}

static void
calls_transfer (Calls *self,
		const gchar *token,
		const gchar *url)
{
  GError *error = NULL;

  dbus_g_proxy_call_no_reply (self->dbus_object, "Transfer",
			      G_TYPE_STRING, token,
			      G_TYPE_STRING, url,
			      G_TYPE_INVALID);
}


static void
update_column (Calls *self, const gchar *token,
	       gint column, const gchar *info)
{
  GtkTreeIter *iter = NULL;
  GtkListStore *store = NULL;

  store =
    GTK_LIST_STORE (gtk_tree_view_get_model (GTK_TREE_VIEW (self->tree)));

  iter = find_iter_for_key (token, 0, store);  

  gtk_list_store_set (store, iter,
		      COLUMN_TOKEN, token,
		      column, info,
		      -1);

  gtk_tree_iter_free (iter);
}

static void connect_entry_activate_callback (GtkWidget *widget, gpointer data)
{
  Calls *self = NULL;

  self = CALLS (data);

  calls_connect (self, gtk_entry_get_text (GTK_ENTRY (widget)));
}

static void transfer_entry_activate_callback (GtkWidget *widget, gpointer data)
{
  Calls *self = NULL;
  GtkTreeSelection *selection = NULL;
  GtkTreeModel *model = NULL;
  GList *paths = NULL;
  GList *ptr = NULL;
  GtkTreeIter iter;
  gchar *token = NULL;
  const gchar *url = NULL;

  self = CALLS (data);
  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (self->tree));
  model = gtk_tree_view_get_model (GTK_TREE_VIEW (self->tree));
  paths = gtk_tree_selection_get_selected_rows (selection, &model);
  url = gtk_entry_get_text (GTK_ENTRY (widget));

  for (ptr = paths ; ptr != NULL; ptr = ptr->next) {

    if (gtk_tree_model_get_iter (model, &iter,(GtkTreePath *)ptr->data)) {

            gtk_tree_model_get (model, &iter, COLUMN_TOKEN, &token, -1);
	    calls_transfer (self, token, url);
	    g_free (token);
    }
  }

  g_list_foreach (paths, (GFunc)gtk_tree_path_free, NULL);
  g_list_free (paths);
}

static void
disconnect_button_callback (GtkWidget *widget, gpointer data)
{
  Calls *self = NULL;
  GtkTreeSelection *selection = NULL;
  GtkTreeModel *model = NULL;
  GList *paths = NULL;
  GList *ptr = NULL;
  GtkTreeIter iter;
  gchar *token = NULL;

  self = CALLS (data);
  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (self->tree));
  model = gtk_tree_view_get_model (GTK_TREE_VIEW (self->tree));
  paths = gtk_tree_selection_get_selected_rows (selection, &model);

  for (ptr = paths ; ptr != NULL; ptr = ptr->next) {

    if (gtk_tree_model_get_iter (model, &iter,(GtkTreePath *)ptr->data)) {

            gtk_tree_model_get (model, &iter, COLUMN_TOKEN, &token, -1);
	    calls_disconnect (self, token);
	    g_free (token);
    }
  }

  g_list_foreach (paths, (GFunc)gtk_tree_path_free, NULL);
  g_list_free (paths);
}

static void
play_pause_button_callback (GtkWidget *widget, gpointer data)
{
  Calls *self = NULL;
  GtkTreeSelection *selection = NULL;
  GtkTreeModel *model = NULL;
  GList *paths = NULL;
  GList *ptr = NULL;
  GtkTreeIter iter;
  gchar *token = NULL;

  self = CALLS (data);
  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (self->tree));
  model = gtk_tree_view_get_model (GTK_TREE_VIEW (self->tree));
  paths = gtk_tree_selection_get_selected_rows (selection, &model);

  for (ptr = paths ; ptr != NULL; ptr = ptr->next) {

    if (gtk_tree_model_get_iter (model, &iter,(GtkTreePath *)ptr->data)) {

            gtk_tree_model_get (model, &iter, COLUMN_TOKEN, &token, -1);
	    calls_play_pause (self, token);
	    g_free (token);
    }
  }

  g_list_foreach (paths, (GFunc)gtk_tree_path_free, NULL);
  g_list_free (paths);
}

static void
update_button_callback (GtkWidget *widget, gpointer data)
{
  Calls *self = CALLS (data);

  calls_flush (self);
  calls_populate (self);
}

static void
state_changed_signal_handler (DBusGProxy *proxy,
			      const char *token,
			      guint state,
			      gpointer data)
{
  calls_update_state (CALLS (data), token, state);
}

static void
name_info_signal_handler (DBusGProxy *proxy,
			  const char *token,
			  const char *name,
			  gpointer data)
{
  update_column (CALLS (data), token, COLUMN_NAME, name);
}

static void
client_info_signal_handler (DBusGProxy *proxy,
			    const char *token,
			    const char *client,
			    gpointer data)
{
  update_column (CALLS (data), token, COLUMN_CLIENT, client);
}

static void
url_info_signal_handler (DBusGProxy *proxy,
			 const char *token,
			 const char *url,
			 gpointer data)
{
  update_column (CALLS (data), token, COLUMN_URL, url);
}

static void protocol_info_signal_handler (DBusGProxy *proxy,
					  const char *token,
					  const char *protocol,
					  gpointer data)
{
  update_column (CALLS (data), token, COLUMN_PROTOCOL, protocol);
}

/* implementation of the public functions */
GtkWidget *
calls_new (void)
{
  Calls *result = NULL;

  result = CALLS (gtk_object_new (CALLS_TYPE,
				  "label", "Calls",
				  NULL));

  return GTK_WIDGET(result);
}
